#include <MarlinArbor.hh>
#include <ArborTool.hh>
#include <ArborToolLCIO.hh>
#include <ArborHit.hh>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/CalorimeterHit.h>
#include <EVENT/SimCalorimeterHit.h>
#include <EVENT/LCFloatVec.h>
#include <EVENT/LCParameters.h>
#include <EVENT/LCGenericObject.h>
#include <IMPL/LCRelationImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimCalorimeterHitImpl.h>
#include <IMPL/CalorimeterHitImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <IMPL/ClusterImpl.h>
#include "UTIL/CellIDDecoder.h"
#include "UTIL/Operators.h"

#include <values.h>
#include <string>
#include <iostream>
#include <cmath>
#include <stdexcept>
#include <TFile.h>
#include <TTree.h>
#include <TMath.h>
#include <Rtypes.h>
#include <sstream>
#include <set>
#include <TVector3.h>
#include <vector>
#include <algorithm>


using namespace std;
using namespace lcio ;
using namespace marlin ;

const string AHCALCellIDDecoder = "M:3,S-1:4,I:9,J:9,K-1:7";    //Same for Ecal and AHCAL
//const string ECALCellIDDecoder  = "M:3,S-1:4,I:9,J:9,K-1:6";	//Need to verify that for DHCAL

const string DHCALCellIDDecoder = "M:3,S-1:4,I:9,J:9,K-1:7";
const string ECALCellIDDecoder = "S-1:3,M:3,K-1:6,I:16,GRZone:3,J:32:16";
CellIDDecoder<CalorimeterHit> idDecoder_ECAL(ECALCellIDDecoder);
CellIDDecoder<CalorimeterHit> idDecoder_DHCAL(DHCALCellIDDecoder);

extern linkcoll InitLinks; 
extern linkcoll IterLinks; 
extern linkcoll links_debug; 
extern branchcoll Trees; 
extern std::vector<int> IsoHitsIndex;

branchcoll alllinks; //To be Merged

std::vector<std::string> CaloHitCollections;

MarlinArbor aMarlinArbor ;
MarlinArbor::MarlinArbor()
        : Processor("MarlinArbor")
          ,
          _output(0)
{
        _description = " Tree Algorithm in Marlin" ;

	std::vector<std::string> EcalCaloHitCollections;
	EcalCaloHitCollections.push_back(std::string("ECALBarrel"));
	EcalCaloHitCollections.push_back(std::string("ECALEndcap"));
	registerInputCollections( LCIO::CALORIMETERHIT,      //adding a flag to Calo /SimCalo
			"EcalHitCollections" ,
			"Hit Collection Names" ,
			_EcalCalCollections ,
			EcalCaloHitCollections);
	
	std::vector<std::string> HcalCaloHitCollections;
        HcalCaloHitCollections.push_back(std::string("HCALBarrel"));
        registerInputCollections( LCIO::CALORIMETERHIT,      //adding a flag to Calo /SimCalo
                        "HcalHitCollections" ,
                        "Hit Collection Names" ,
                        _HcalCalCollections ,
                        HcalCaloHitCollections);
}

void MarlinArbor::init() {

	printParameters();

	for(unsigned int i0 = 0; i0 < _EcalCalCollections.size(); i0++)
	{
		CaloHitCollections.push_back(_EcalCalCollections[i0].c_str());
	}
	for(unsigned int j0 = 0; j0 < _HcalCalCollections.size(); j0++)
	{
		CaloHitCollections.push_back(_HcalCalCollections[j0].c_str());
	}

}

void MarlinArbor::HitsPreparation()
{
	cout<<"Start to prepare Hits"<<endl;
}

void MarlinArbor::LinkVisulization( LCEvent * evtPP, std::string Name, std::vector<CalorimeterHit*> Hits, linkcoll inputLinks ) 
{

	LCCollectionVec *colllink = new LCCollectionVec(LCIO::LCRELATION);
	LCFlagImpl linkflag;
	linkflag.setBit(LCIO::CHBIT_LONG);
	colllink->setFlag(linkflag.getFlag());

	int NLink = inputLinks.size();
	std::pair<int, int> currlink; 
	for(int i0 = 0; i0 < NLink; i0++)
	{
		currlink = inputLinks[i0];

		CalorimeterHit* a_hit = Hits[currlink.first];
		CalorimeterHit* b_hit = Hits[currlink.second];

		LCRelationImpl *a_link = new LCRelationImpl(a_hit, b_hit);
		colllink->addElement( a_link );
	}

	evtPP->addCollection(colllink, Name);
}

void MarlinArbor::MakeBush( LCEvent * evtPP, std::vector<std::string> inputTreeCollections, std::string outputBushCollection )
{
	LCCollection *currbushcoll = new LCCollectionVec(LCIO::CLUSTER);
	LCFlagImpl bushflag;
	bushflag.setBit(LCIO::CHBIT_LONG);
	currbushcoll->setFlag(bushflag.getFlag());	

	int Ncoll = inputTreeCollections.size();
	int NTree = 0;
	for(int i = 0; i < Ncoll; i++)
	{
		LCCollection * a_TreeColl = evtPP ->getCollection(inputTreeCollections[i].c_str());
		NTree = a_TreeColl->getNumberOfElements();
		for(int j = 0; j < NTree; j++)
		{
			Cluster * a_tree = dynamic_cast<Cluster*>(a_TreeColl->getElementAt(j));
			ClusterImpl *a_bush = NaiveCluImpl(a_tree);
			currbushcoll->addElement(a_bush);
		}
	}

	evtPP->addCollection( currbushcoll, outputBushCollection );
}

void MarlinArbor::MakeIsoHits( LCEvent * evtPP, std::vector<CalorimeterHit*> inputCaloHits, std::string outputBushCollection )
{
	LCCollection * isohitcoll = new LCCollectionVec(LCIO::CALORIMETERHIT);

	string initString = "M:3,S-1:3,I:9,J:9,K-1:6";          //Need to verify
	isohitcoll->parameters().setValue(LCIO::CellIDEncoding, initString);

	LCFlagImpl flag;
	flag.setBit(LCIO::CHBIT_LONG);                  
	flag.setBit(LCIO::CHBIT_ID1);
	flag.setBit(LCIO::RCHBIT_ENERGY_ERROR);
	isohitcoll->setFlag(flag.getFlag());

	int nhit = inputCaloHits.size();

	for(int i = 0; i < nhit; i++)
	{
		CalorimeterHit* a_hit = inputCaloHits[i];
		CalorimeterHitImpl * collhit = new CalorimeterHitImpl();
		collhit->setPosition(a_hit->getPosition());
		collhit->setCellID0(a_hit->getCellID0());
		collhit->setCellID1(a_hit->getCellID1());
		collhit->setEnergy(a_hit->getEnergy());
		isohitcoll->addElement(collhit);
	}

	evtPP->addCollection(isohitcoll, outputBushCollection);
}

void MarlinArbor::MakeMergedBush( LCEvent * evtPP, std::vector<std::string> inputTreeCollections, std::string outputBushCollection, float TreeSeedDisThreshold )
{
	LCCollection *currbushcoll = new LCCollectionVec(LCIO::CLUSTER);
	LCCollection *currbushes = new LCCollectionVec(LCIO::CLUSTER);
	LCFlagImpl bushflag;
	bushflag.setBit(LCIO::CHBIT_LONG);
	currbushcoll->setFlag(bushflag.getFlag());
	currbushes->setFlag(bushflag.getFlag());

	int Ncoll = inputTreeCollections.size();
	int NTree = 0;
	for(int i = 0; i < Ncoll; i++)
	{
		LCCollection * a_TreeColl = evtPP ->getCollection(inputTreeCollections[i].c_str());
		NTree = a_TreeColl->getNumberOfElements();
		for(int j = 0; j < NTree; j++)
		{
			Cluster * a_tree = dynamic_cast<Cluster*>(a_TreeColl->getElementAt(j));
			ClusterImpl *a_bush = NaiveCluImpl(a_tree);
			currbushcoll->addElement(a_bush);
		}
	}

	int NtotalTree = currbushcoll->getNumberOfElements();;
	Cluster* a_tree, *b_tree; 

	TVector3 PosA, PosB; // PosB, PosAB; 
	TVector3 PosEndA, PosEndB;
	float Angle1, Angle2, Angle3, Angle4;
	float DisAProjectB = 0;
	float DisBProjectA = 0;
	float DisBushAB = 0;
	int SizeA = 0; 
	int SizeB = 0;
	int NJoints = 0; 
	float DisSeedBush = 1.0E9;

	TMatrixF FlagTreeMerge(NtotalTree, NtotalTree);

	for(int i0 = 0; i0 < NtotalTree; i0++)
	{
		a_tree = dynamic_cast<Cluster*>( currbushcoll->getElementAt(i0));
		PosA = a_tree->getPosition();
		PosEndA = CluEndP(a_tree);
		SizeA = a_tree->getCalorimeterHits().size();

		for(int j0 = i0 + 1; j0 < NtotalTree; j0++)	
		{
			b_tree = dynamic_cast<Cluster*>( currbushcoll->getElementAt(j0));
			PosB = b_tree->getPosition();
			PosEndB = CluEndP(b_tree);
			SizeB = b_tree->getCalorimeterHits().size();
			DisSeedBush = TMath::Min(DisPointToBush(PosA, b_tree), DisPointToBush(PosB, a_tree));
			NJoints = JointsBetweenBush(a_tree, b_tree, 14);

			Angle1 = (PosEndA - PosA).Angle(PosA - PosEndB);        //A project on B End
			Angle2 = (PosA - PosEndB).Angle(PosEndB - PosB);
			Angle3 = (PosEndB - PosB).Angle(PosB - PosEndA);
			Angle4 = (PosB - PosEndA).Angle(PosEndA - PosA);

			DisBushAB = BushDis(a_tree, b_tree);
			DisAProjectB = CluProjectiveDis(PosEndA, PosEndA - PosA, b_tree);
			DisBProjectA = CluProjectiveDis(PosEndB, PosEndB - PosB, a_tree);

			if(DisSeedBush < TreeSeedDisThreshold ) // || NJoints > 0.2*(TMath::Min(SizeA, SizeB)))
			{
				FlagTreeMerge(j0, i0) = 1;
				FlagTreeMerge(i0, j0) = 1;
			}
			else if( (SizeA < 8 || SizeB < 8) && NJoints > 2 )
			{
				FlagTreeMerge(j0, i0) = 2;
				FlagTreeMerge(i0, j0) = 2;
			}
			else if( DisBushAB < 40 && ( DisAProjectB < 2 + i0*25 || DisBProjectA < 2 + i0*25 ) )
			{
				FlagTreeMerge(j0, i0) = 0;
				FlagTreeMerge(i0, j0) = 0;
			}
			else if( PosEndB.Mag() < PosA.Mag() && (PosEndB - PosA).Mag() < 200 )
			{
				if( Angle1 < 0.2 )
				{
					FlagTreeMerge(j0, i0) = 3;
					FlagTreeMerge(i0, j0) = 3;
				}
				else if( Angle2 < 0.2)
				{
					FlagTreeMerge(j0, i0) = 4;
					FlagTreeMerge(i0, j0) = 4;
				}
			}
			else if( PosEndA.Mag() < PosB.Mag() && (PosEndA - PosB).Mag() < 200 )
			{
				if( Angle3 < 0.2 )
				{
					FlagTreeMerge(j0, i0) = 5;
					FlagTreeMerge(i0, j0) = 5;
				}
				else if( Angle4 < 0.2 )
				{
					FlagTreeMerge(j0, i0) = 6;
					FlagTreeMerge(i0, j0) = 6;
				}
			}
		}	
	}

	TMatrixF TreeMergeSYM = MatrixSummarize(FlagTreeMerge);
	currbushes = ClusterMerge(currbushcoll, TreeMergeSYM);

	evtPP->addCollection( currbushes, outputBushCollection );
}


void MarlinArbor::processEvent( LCEvent * evtP )
{
	if(evtP)
	{

		int EvtNr = evtP->getEventNumber();
		_eventNr = EvtNr; 

		std::cout<<EvtNr<<" events processed"<<std::endl;

		MarlinArbor::HitsPreparation();	//Absorb isolated hits; 

		TVector3 currHitPos;

		std::vector< TVector3 > inputHitsPos;
		std::vector< ArborHit > inputABHit; 
		std::vector< CalorimeterHit* > inputHits;  
		std::vector< std::vector<int> > Sequence; 
		std::vector< int > currBranch; 
		std::string BranchStringName, LinkStringName, LinkStringNameSS, TreeStringName, DebugLinkString;
		std::vector<float> HitMag; 
		std::vector<int> HitOrder; 
		std::vector<std::string> EcalTrees; 
		std::vector<std::string> HcalTrees;
		EcalTrees.clear();
		HcalTrees.clear();
		alllinks.clear();
		float cellsize = 0; 
		float layerthick = 0;
		int DHCALFlag = 0; 
		int IsoHitSize = 0; 
		int LayerNum = 0; 
		int StaveNum = 0; 

		std::vector<CalorimeterHit*> EcalIsoHits;
		std::vector<CalorimeterHit*> HcalIsoHits;

		for(unsigned int i1 = 0; i1 < CaloHitCollections.size(); i1++)
		{

			try{	//not necessary to run for every event. anyway...
				LCCollection * CaloHitColl = evtP ->getCollection(CaloHitCollections[i1].c_str());
				BranchStringName = "Branch_";
				BranchStringName += CaloHitCollections[i1].c_str();
				LinkStringName = "Links_";
				LinkStringName += CaloHitCollections[i1].c_str();
				LinkStringNameSS = "LinksIterAll_";
				LinkStringNameSS += CaloHitCollections[i1].c_str();
				TreeStringName = "Tree_";
				TreeStringName += CaloHitCollections[i1].c_str();
				DebugLinkString = "LinksDebug_";
				DebugLinkString += CaloHitCollections[i1].c_str();

				DHCALFlag = 0;
				/*
				   if( i1 < _EcalCalCollections.size() )		//Be very careful!
				   DHCALFlag = 0;  
				   else 
				   DHCALFlag = 1; 
				   */

				inputHitsPos.clear();
				inputABHit.clear();
				inputHits.clear();
				Sequence.clear();
				HitMag.clear();
				HitOrder.clear();

				int NHitsCurrCol = CaloHitColl->getNumberOfElements();

				CellIDDecoder<CalorimeterHit> idDecoder(CaloHitColl);

				if(i1 < _EcalCalCollections.size()) 
				{
					cellsize = 5; 
					layerthick = 8;	//well...	//default 8
				}
				else 
				{
					cellsize = 10;		// 20
					layerthick = 26.5; 	//26.5
				}

				// int NLayer = 1; 
				// double HitTime = 0.12345; 

				for(int j1 = 0; j1 < NHitsCurrCol; j1++)
				{
					CalorimeterHit *a_Digihit = dynamic_cast<CalorimeterHit*>( CaloHitColl->getElementAt( j1 ) );
					currHitPos = a_Digihit->getPosition();
					HitMag.push_back( currHitPos.Mag() );
				} 
				HitOrder = SortMeasure(HitMag, 1);      //From Outer to Inner
				for(int k1 = 0; k1 < NHitsCurrCol; k1++)
				{
					CalorimeterHit *b_Digihit = dynamic_cast<CalorimeterHit*>( CaloHitColl->getElementAt( HitOrder[k1] ) );
					currHitPos = b_Digihit->getPosition();

					if( CaloHitCollections[i1] == "LCAL")
					{
						LayerNum = idDecoder(b_Digihit)["K"];
						StaveNum = 0;
					}
					else
					{
						LayerNum = idDecoder(b_Digihit)["K-1"] + 1;
						StaveNum = idDecoder(b_Digihit)["S-1"] + 1;
					}
					ArborHit a_abhit(currHitPos, LayerNum, 0, StaveNum, i1);	//Last Number is the Hit time

					// cout<<CaloHitCollections[i1]<<"A_ABHIT "<<a_abhit.GetTime()<<" : "<<(a_abhit.GetPosition()).X()<<" : "<<(a_abhit.GetPosition()).Y()<<" : "<<(a_abhit.GetPosition()).Z()<<"  :  "<<a_abhit.GetLayer()<<endl; 

					inputABHit.push_back(a_abhit);
					// inputHitsPos.push_back( currHitPos );
					inputHits.push_back(b_Digihit);
				}

				// Sequence = Arbor(inputHitsPos, cellsize, layerthick);
				Sequence = Arbor(inputABHit, cellsize, layerthick);

				IsoHitSize = IsoHitsIndex.size();

				for(int i2 = 0; i2 < IsoHitSize; i2++)
				{
					CalorimeterHit *a_IsoHit = dynamic_cast<CalorimeterHit*>( CaloHitColl->getElementAt(HitOrder[IsoHitsIndex[i2]]) );
					if( i1 < _EcalCalCollections.size() )
						EcalIsoHits.push_back(a_IsoHit);
					else
						HcalIsoHits.push_back(a_IsoHit);
				}

				if(i1 < _EcalCalCollections.size())
				{
					EcalTrees.push_back(TreeStringName);
				}
				else
				{
					HcalTrees.push_back(TreeStringName);
				}

				for(unsigned int s1 = 0; s1 < Sequence.size(); s1++)
				{
					std::vector< int > a_branch = Sequence[s1];
					alllinks.push_back(a_branch);
				}

				cout<<i1<<"th Collection "<<CaloHitCollections[i1].c_str()<<" has Size "<< Sequence.size() <<" all link Size "<<alllinks.size()<<endl; 

				ClusterBuilding(evtP, BranchStringName, inputHits, Sequence, DHCALFlag);

				ClusterBuilding(evtP, TreeStringName, inputHits, Trees, DHCALFlag);

				LinkVisulization(evtP, LinkStringName, inputHits, InitLinks);

				LinkVisulization(evtP, LinkStringNameSS, inputHits, IterLinks);

				LinkVisulization(evtP, DebugLinkString, inputHits, links_debug);

			}catch  (lcio::DataNotAvailableException zero) { }

		}

		std::string EcalBush = "EcalBushes";
		std::string HcalBush = "HcalBushes";
		std::string EcalIsoHitName = "EcalIsoHits_1";
		std::string HcalIsoHitName = "HcalIsoHits_1";

		MakeBush(evtP, EcalTrees, EcalBush); //, 20);
		MakeBush(evtP, HcalTrees, HcalBush); //, 40);

		MakeIsoHits(evtP, EcalIsoHits, EcalIsoHitName);
		MakeIsoHits(evtP, HcalIsoHits, HcalIsoHitName);		// New Pattern?


		LCCollection * EcalIsoHitsColl = evtP ->getCollection( "EcalIsoHits_1" );
		LCCollection * HcalIsoHitsColl = evtP ->getCollection( "HcalIsoHits_1" );
		// inputHitsPos.clear();
		inputABHit.clear();
		inputHits.clear();	
		for(int i3 = 0; i3 < EcalIsoHitsColl->getNumberOfElements(); i3++)
		{
			CalorimeterHit *b_Digihit = dynamic_cast<CalorimeterHit*>( EcalIsoHitsColl->getElementAt( i3 ) );
			currHitPos = b_Digihit->getPosition();
			// inputHitsPos.push_back( currHitPos );
			LayerNum = (idDecoder_ECAL(b_Digihit)["K-1"]+1);
			StaveNum = (idDecoder_ECAL(b_Digihit)["S-1"]+1);
			ArborHit a_abhit(currHitPos, LayerNum, 0, StaveNum, 11);
			inputABHit.push_back(a_abhit);
			inputHits.push_back(b_Digihit);
		}
		Sequence = Arbor(inputABHit, 40, 20);	//Re-clustering with 8 times larger distance		
		ClusterBuilding( evtP, "Ecal_IterFrag", inputHits, Trees, 0 );

		IsoHitSize = IsoHitsIndex.size();

		EcalIsoHits.clear();
		for(int i2 = 0; i2 < IsoHitSize; i2++)
		{
			CalorimeterHit *a_IsoHit = dynamic_cast<CalorimeterHit*>( EcalIsoHitsColl->getElementAt(IsoHitsIndex[i2] ));
			EcalIsoHits.push_back(a_IsoHit);
		}

		inputABHit.clear();
		inputHits.clear();
		for(int i4 = 0; i4 < HcalIsoHitsColl->getNumberOfElements(); i4++)
		{
			CalorimeterHit *b_Digihit = dynamic_cast<CalorimeterHit*>( HcalIsoHitsColl->getElementAt( i4 ) );
			currHitPos = b_Digihit->getPosition();
			LayerNum = (idDecoder_DHCAL(b_Digihit)["K-1"]+1);
			StaveNum = (idDecoder_DHCAL(b_Digihit)["S-1"]+1);
			ArborHit a_abhit(currHitPos, LayerNum, 0, StaveNum, 12);
			inputABHit.push_back(a_abhit);
			inputHits.push_back(b_Digihit);
		}
		Sequence = Arbor(inputABHit, 50, 60); //Re-clustering with 4 times larger distance            
		ClusterBuilding( evtP, "Hcal_IterFrag", inputHits, Trees, 0 );

		IsoHitSize = IsoHitsIndex.size();

		HcalIsoHits.clear();
		for(int i4 = 0; i4 < IsoHitSize; i4++)
		{
			CalorimeterHit *a_IsoHit = dynamic_cast<CalorimeterHit*>( HcalIsoHitsColl->getElementAt(IsoHitsIndex[i4]) );
			HcalIsoHits.push_back(a_IsoHit);
		}

		MakeIsoHits(evtP, EcalIsoHits, "EcalIsoHits");
		MakeIsoHits(evtP, HcalIsoHits, "HcalIsoHits");

	}
}


void MarlinArbor::end()
{
	std::cout<<"Arbor Ends. Good luck"<<std::endl;
}
